import numpy as np
from matplotlib import pyplot as plt
from sympy import Q
from fwk.coloring import ccolors

# 3D spline with Bézier curves
from scipy.interpolate import bisplrep
from scipy.interpolate import bisplev
from scipy.interpolate import splrep
from scipy.interpolate import splev

from scipy.interpolate import interp1d

from scipy.interpolate import RBFInterpolator

from scipy.interpolate import interp2d

import fwk


class Interpolator:
    def __init__(self, _dartAPI, _blw, _imsh, _vmsh, _cfg):
        
        """fromBlw = np.zeros((0,5))
        for e in _blw[1].tag.elems:
            if e.ptag.name == "wake" or e.ptag.name == "wake_":
                for n in e.nodes:
                    fromBlw = np.row_stack((fromBlw, [n.pos[0], n.pos[1], n.pos[2], n.row, e.no]))

        fromiMsh = np.zeros((0,5))
        for e in _imsh.elems:
            if e.ptag.name == "wake" or e.ptag.name == "wake_":
                for n in e.nodes:
                    fromiMsh = np.row_stack((fromiMsh, [n.pos[0], n.pos[1], n.pos[2], n.row, e.no]))

        for i in range(len(fromBlw[:,0])):
            if fromBlw[i, :].any() != fromiMsh[i, :].any():
                print(i)
                print(fromBlw[i, :])
                print(fromiMsh[i, :])
            else:
                print('ok', i)
        quit()"""
        self.cfg = _cfg
        symmetryBC = self.cfg['Sym']

        self.tms = fwk.Timers()

        self.tms['Mesh sort'].start()
        self.idata, self.icg = self.GetMesh(_imsh)
        self.baseiData = self.idata.copy()
        for sym in symmetryBC:
            for iReg in range(len(self.idata)):
                dummy = self.baseiData[iReg].copy()
                dummy[:,1] -= sym
                dummy[:,1] *= -1
                dummy[:,1] += sym
                self.idata[iReg] = np.row_stack((self.idata[iReg], dummy))
        self.vdata, self.vcg = self.GetMesh(_vmsh, visc=1)
        self.Xp, self.Yp, self.Zp, self.sortedRows, self.Xc, self.Yc, self.Zc, self.sortedElems = self.SortMesh(self.vdata, self.vcg)
        self.tms['Mesh sort'].stop()
        print(self.tms)

        """fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        ax.scatter(self.vdata[0][:,0], self.vdata[0][:,1], self.vdata[0][:,2], s=0.3)
        ax.scatter(self.vcg[0][:,0], self.vcg[0][:,1], self.vcg[0][:,2], s=0.3)
        ax.scatter(self.vdata[1][:,0], self.vdata[1][:,1], self.vdata[1][:,2], s=0.3)
        ax.scatter(self.vcg[1][:,0], self.vcg[1][:,1], self.vcg[1][:,2], s=0.3)
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        #plt.gca().invert_yaxis()
        plt.axis('off')
        plt.show()

        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        ax.scatter(self.idata[0][:,0], self.idata[0][:,1], self.idata[0][:,2], s=0.3)
        ax.scatter(self.icg[0][:,0], self.icg[0][:,1], self.icg[0][:,2], s=0.3)
        ax.scatter(self.idata[1][:,0], self.idata[1][:,1], self.idata[1][:,2], s=0.3)
        ax.scatter(self.icg[1][:,0], self.icg[1][:,1], self.icg[1][:,2], s=0.3)
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        #plt.gca().invert_yaxis()
        #plt.axis('off')
        plt.show()"""
        
        self.iSolver = _dartAPI
        self.blw = _blw

        # Create data structures for interpolation
        self.tms = fwk.Timers()

        print(self.tms)

        self.stgPt=np.zeros(len(self.cfg['Sections']), dtype=int)

        self.xxPrev = [[np.zeros(0) for iReg in range(3)] for iSec in range(len(self.cfg['Sections']))]
        #self.uePrev = [[np.zeros(len(self.viscStruct[iSec][iReg][:,0])) for iReg in range(3)] for iSec in range(len(self.viscStruct))]
        self.DeltaStarPrev = [[np.zeros(0) for iReg in range(3)] for iSec in range(len(self.cfg['Sections']))]

        self.Cg_Sec = [[np.zeros((len(self.Xc[iReg]), 3)) for iReg in range(2)] for _ in range(len(self.cfg['Sections']))]

        self.xCp = [[np.zeros((len(self.Xc[iReg]), 1)) for iReg in range(2)] for _ in range(len(self.cfg['Sections']))]
        self.CpIter = [[np.zeros((len(self.Xc[iReg]), 1)) for iReg in range(2)] for _ in range(len(self.cfg['Sections']))]

    def GetInvBC(self, vSolver, couplIter):
        self.tms['InvToVisc'].start()

        # Extract parameters from the inviscid solver
        # Particular care must be taken with space directions; In the inviscid, the airfoil is in the x,z plane.
        # While setting up the viscous solver, y and z must be inverted. 

        # Wing

        U = [np.zeros((len(self.idata[i][:,0]), 3)) for i in range(len(self.idata))]
        M = [np.zeros(len(self.idata[i][:,0])) for i in range(len(self.idata))]
        Rho = [np.zeros(len(self.idata[i][:,0])) for i in range(len(self.idata))]
        # Extract variables.

        for iReg in range(len(self.idata)):
            for iNode, row in enumerate(self.idata[iReg][:,3]):
                row = int(row)
                U[iReg][iNode,:] = [self.iSolver.U[row][0], self.iSolver.U[row][1], self.iSolver.U[row][2]]
                M[iReg][iNode] = self.iSolver.M[row]
                Rho[iReg][iNode] = self.iSolver.rho[row]
        
        #self.URani = U

        nD = self.cfg['nDim']
        Ux = [self.__RbfToSections(self.idata[0][:,:nD], U[0][:,0], self.vdata[0][:,:nD]), self.__RbfToSections(self.idata[1][:, 0:2], U[1][:,0], self.vdata[1][:,0:2])]
        Uy = [self.__RbfToSections(self.idata[0][:,:nD], U[0][:,1], self.vdata[0][:,:nD]), self.__RbfToSections(self.idata[1][:, 0:2], U[1][:,1], self.vdata[1][:,0:2])]
        Uz = [self.__RbfToSections(self.idata[0][:,:nD], U[0][:,2], self.vdata[0][:,:nD]), self.__RbfToSections(self.idata[1][:, 0:2], U[1][:,2], self.vdata[1][:,0:2])]
        Me = [self.__RbfToSections(self.idata[0][:,:nD], M[0], self.vdata[0][:,:nD]), self.__RbfToSections(self.idata[1][:, 0:2], M[1], self.vdata[1][:,0:2])]
        Rhoe = [self.__RbfToSections(self.idata[0][:,:nD], Rho[0], self.vdata[0][:,:nD]), self.__RbfToSections(self.idata[1][:, 0:2], Rho[1], self.vdata[1][:,0:2])]
        
        #self.URani = Ux[0]
        for iSec, y in enumerate(self.cfg['Sections']):

            for iReg in range(2):
                
                # Y Z inverted
                jcol = (abs(self.Yp[iReg][0,:]-y)).argmin()
                x_pos = self.Xp[iReg][:,jcol]
                y_pos = self.Yp[iReg][:,jcol]
                z_pos = self.Zp[iReg][:,jcol]
                goodRows = self.sortedRows[iReg][:,jcol]

                ux = np.zeros(0)
                uy = np.zeros(0)
                uz = np.zeros(0)
                me = np.zeros(0)
                rhoe = np.zeros(0)
                cp = np.zeros(0)
                xxx = np.zeros(0)

                for i in range(len(x_pos)):
                    # Y Z inverted 
                    row = np.where(self.vdata[iReg][:,3]==goodRows[i])[0]
                    xxx = np.append(xxx, self.vdata[iReg][row,0])
                    ux = np.append(ux, Ux[iReg][row])
                    uy = np.append(uy, Uy[iReg][row])
                    uz = np.append(uz, Uz[iReg][row])
                    me = np.append(me, Me[iReg][row])
                    rhoe = np.append(rhoe, Rhoe[iReg][row])

                #self.xCp[iSec][iReg] = np.column_stack((self.xCp[iSec][iReg], abs(x_pos)))
                #self.CpIter[iSec][iReg] = np.column_stack((self.CpIter[iSec][iReg], cp))
                
                if couplIter == 0:

                    if iReg == 0:
                        self.stgPt[iSec] = np.argmin(ux**2+uz**2)
                        xUp, xLw = self.__Remesh(abs(x_pos), self.stgPt[iSec])
                        yUp, yLw = self.__Remesh(y_pos, self.stgPt[iSec])
                        zUp, zLw = self.__Remesh(z_pos, self.stgPt[iSec])

                        vSolver.SetGlobMesh(iSec, 0, xUp, zUp, yUp)
                        vSolver.SetGlobMesh(iSec, 1, xLw, zLw, yLw)

                        self.DeltaStarPrev[iSec][0] = np.zeros(len(xUp))
                        self.DeltaStarPrev[iSec][1] = np.zeros(len(xLw))

                        self.xxPrev[iSec][0] = np.zeros(len(xUp))
                        self.xxPrev[iSec][1] = np.zeros(len(xLw))
                    else:
                        vSolver.SetGlobMesh(iSec, 2, x_pos, z_pos, y_pos)
                        self.DeltaStarPrev[iSec][2] = np.zeros(len(x_pos))   
                        self.xxPrev[iSec][2] = np.zeros(len(x_pos))
                    
                    for i in range(len(x_pos)-1):
                        self.Cg_Sec[iSec][iReg][i,0] = x_pos[i+1] + x_pos[i]
                        self.Cg_Sec[iSec][iReg][i,1] = y_pos[i+1] + y_pos[i]
                        self.Cg_Sec[iSec][iReg][i,2] = z_pos[i+1] + z_pos[i]
                    self.Cg_Sec[iSec][iReg]/=2

                if iReg == 0:
                    UxUp, UxLw = self.__Remesh(ux, self.stgPt[iSec])
                    UyUp, UyLw = self.__Remesh(uy, self.stgPt[iSec])
                    UzUp, UzLw = self.__Remesh(uz, self.stgPt[iSec])

                    vSolver.SetVelocities(iSec, 0, UxUp, UzUp, UyUp)
                    vSolver.SetVelocities(iSec, 1, UxLw, UzLw, UyLw)

                    MeUp, MeLw = self.__Remesh(me, self.stgPt[iSec])
                    vSolver.SetMe(iSec, 0, MeUp)
                    vSolver.SetMe(iSec, 1, MeLw)

                    RhoUp, RhoLw = self.__Remesh(rhoe, self.stgPt[iSec])
                    vSolver.SetRhoe(iSec, 0, RhoUp)
                    vSolver.SetRhoe(iSec, 1, RhoLw)
                else:

                    vSolver.SetVelocities(iSec, 2, ux, uz, uy)
                    vSolver.SetMe(iSec, 2, me)
                    vSolver.SetRhoe(iSec, 2, rhoe)

            for iViscR in range(3):
                vSolver.SetDeltaStarExt(iSec, iViscR, self.DeltaStarPrev[iSec][iViscR])
                vSolver.SetxxExt(iSec, iViscR, self.xxPrev[iSec][iViscR])

    def SetBlowingVel(self, vSolver, couplIter):

        for iSec in range(len(self.cfg['Sections'])):
            for iViscR in range(3):
                self.DeltaStarPrev[iSec][iViscR] = vSolver.ExtractDeltaStar(iSec, iViscR)
                self.xxPrev[iSec][iViscR] = vSolver.Extractxx(iSec, iViscR)

        blw = [[np.zeros(0) for _ in range(2)] for _ in range(len(self.cfg['Sections']))]

        for iSec in range(len(self.cfg['Sections'])):
            for iReg in range(2):
                if iReg == 0:
                    blwUp = vSolver.ExtractBlowingVel(iSec, 0)
                    blwLw = vSolver.ExtractBlowingVel(iSec, 1)
                    blw[iSec][iReg] = np.concatenate((blw[iSec][iReg], blwUp[::-1], blwLw))
                elif iReg == 1:
                    blw[iSec][iReg] = np.concatenate((blw[iSec][iReg], vSolver.ExtractBlowingVel(iSec, 2)))
        
        for iReg in range(2):

            # Find index of all Sections in Y

            jBounds = np.zeros(len(self.cfg['Sections']), dtype=int)
            for i, y in enumerate(self.cfg['Sections']):
                jBounds[i] = (abs(self.Yp[iReg][0,:]-y)).argmin()

            # Find all index of column of cg which lies between sections
            # len of indexInside = nb of intervals between sections (nSections-1)

            indexInside = [[] for _ in range(len(self.cfg['Sections'])-1)]

            for i in range(len(indexInside)):
                
                indexInside[i] = np.where((self.Yp[iReg][0,jBounds[i]] <= self.Yc[iReg][0,:]) & (self.Yc[iReg][0,:] <= self.Yp[iReg][0,jBounds[i+1]]))[0]

            indexOutside = [np.where((self.Yc[iReg][0,:] <= self.Yp[iReg][0,jBounds[0]]))[0], np.where((self.Yc[iReg][0,:] >= self.Yp[iReg][0,jBounds[-1]]))[0]]

            blwAll = np.zeros((len(self.Xc[iReg][:,0]), len(self.Xc[iReg][0,:])))

            dummy = np.zeros((len(self.Xc[iReg][:,0]), len(indexOutside[0])))
            for iCol in range(len(dummy[0,:])):
                dummy[:,iCol] = blw[0][iReg]
                
            blwAll[:,indexOutside[0]] = dummy

            for iInter, inter in enumerate(indexInside):
                dblw = blw[iInter+1][iReg] - blw[iInter][iReg]
                dy = self.Cg_Sec[iInter+1][iReg][0,1] - self.Cg_Sec[iInter][iReg][0,1]
                for j in inter:
                    if dy == 0:
                        print('dy=0 encountered. mem1 = ',self.Cg_Sec[iInter+1][iReg][0,1], 'mem2 = ',self.Cg_Sec[iInter][iReg][0,1])
                        print(iInter)
                        quit()
                    blwAll[:, j] = dblw/dy*(self.Yc[iReg][0, j] - self.Cg_Sec[iInter][iReg][0,1]) + blw[iInter][iReg]

            dummy = np.zeros((len(self.Xc[iReg][:,0]), len(indexOutside[1])))
            for iCol in range(len(dummy[0,:])):
                dummy[:,iCol] = blw[-1][iReg]
                
            blwAll[:,indexOutside[1]] = dummy

            """fig = plt.figure()
            ax = fig.add_subplot(projection='3d')
            for i in range(len(blw)):
                ax.scatter(self.Cg_Sec[i][iReg][:,0], self.Cg_Sec[i][iReg][:,1], blw[i][iReg], s=0.7, color='red')
            for i in range(len(blwAll[0,:])):
                ax.scatter(self.Xc[iReg][:,i], self.Yc[iReg][:,i], blwAll[:,i], s=0.2, color='blue')
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            #plt.gca().invert_yaxis()
            plt.show()"""
            
            nD = self.cfg['nDim']
            xyz = np.zeros((np.size(self.Xc[iReg]), 3))
            blwForMap = np.zeros(np.size(self.Xc[iReg]))
            ind=0
            for i in range(len(self.Xc[iReg][:,0])):
                for j in range(len(self.Xc[iReg][0,:])):
                    xyz[ind,:] = [self.Xc[iReg][i,j], self.Yc[iReg][i,j], self.Zc[iReg][i,j]]
                    blwForMap[ind] = blwAll[i,j]
                    ind+=1
            
            if iReg == 0:
                mappedBlw = self.__RbfToSections(xyz, blwForMap, self.icg[iReg][:, 0:3])
            elif iReg == 1:
                mappedBlw = self.__RbfToSections(xyz[:,0:2], blwForMap, self.icg[iReg][:, 0:2])

            
            """fig = plt.figure()
            ax = fig.add_subplot(projection='3d')
            for i in range(len(blw)):
                ax.scatter(self.Cg_Sec[i][iReg][:,0], self.Cg_Sec[i][iReg][:,1], blw[i][iReg], s=0.7, color='red')
            ax.scatter(self.icg[iReg][:, 0], self.icg[iReg][:, 1], mappedBlw, s=0.2, color='blue')
            #ax.set_zlim([-0.01, 0.015])
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')
            #plt.gca().invert_yaxis()
            plt.show()"""


            self.tms['Interpolation'].stop()
            if iReg == 0:
                self.URani = mappedBlw
            if iReg==0 or iReg==1:
                for iElm, e in enumerate(self.blw[iReg].tag.elems):
                    if np.size(mappedBlw[self.icg[iReg][:,3]==e.no])>1:
                        print(iElm, mappedBlw[self.icg[iReg][:,3]==e.no])
                        raise RuntimeError('Too many elements found matching the element number')
                    self.blw[iReg].setU(iElm, mappedBlw[self.icg[iReg][:,3]==e.no][0])

    def RunSolver(self):
        self.iSolver.run()

    def ResetSolver(self):
        self.iSolver.reset()

    def GetCp(self,bnd):
        import dart.utils as floU
        Cp = floU.extract(bnd, self.iSolver.Cp)
        return Cp

    def GetAlpha(self):
        return self.iSolver.pbl.alpha

    def GetMinf(self):
        return self.iSolver.pbl.M_inf
    
    def GetCl(self):
        return self.iSolver.Cl
    
    def GetCm(self):
        return self.iSolver.Cm

    def __Remesh(self, vec, stgPt):
        xUp = vec[:stgPt+1]
        xUp = xUp[::-1]

        xLw = vec[stgPt:]
        return xUp, xLw

    def __RbfToSections(self, x, var, xs):
        if np.all(var == var[0]):
            varOut = RBFInterpolator(x, var, neighbors=1, kernel='linear', smoothing=0.0, degree=0)(xs)
        else:
            varOut = RBFInterpolator(x, var, neighbors=self.cfg['neighbors'], kernel=self.cfg['rbftype'], smoothing=self.cfg['smoothing'], degree=self.cfg['degree'])(xs)
        return varOut


    def __GetWing(self, blw, config):

        viscStruct = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0]-1, 3)) for i in range(2)] for _ in range(config['nSections'])]
        viscStruct_tr = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0]-1, 3)) for i in range(2)] for _ in range(config['nSections'])]

        
        invStruct = [np.zeros((len(blw[i].nodes), 4)) for i in range(len(blw))]
        invStruct_tr = [np.zeros((len(blw[i].nodes), 4)) for i in range(len(blw))]

        # Wing
        upperNodes = np.zeros((0,4))
        lowerNodes = np.zeros((0,4))

        self.tms['ElmLoop'].start()
        # Compute reference normal
        nVec = [np.zeros(0) for _ in range(blw[1].tag.elems[0].nodes.size())]
        for iNode, n in enumerate(blw[1].tag.elems[0].nodes):
            nVec[iNode] = np.array([n.pos[0], n.pos[1], n.pos[2]])
        normalRef = np.cross(nVec[1] - nVec[0], nVec[2] - nVec[0])
        normalRef/=np.linalg.norm(normalRef)
            
        for e in blw[0].tag.elems:
            nVec = [np.zeros(0) for _ in range(e.nodes.size())]
            for iNode, n in enumerate(e.nodes):
                nVec[iNode] = np.array([n.pos[0], n.pos[1], n.pos[2]])
            normal = np.cross(nVec[1] - nVec[0], nVec[2] - nVec[0])
            normal/=np.linalg.norm(normalRef)
            if np.dot(normal, normalRef) >= 0:
                for n in e.nodes:
                    if n.row not in upperNodes[:,3] and n.row not in lowerNodes[:,3]:
                        upperNodes = np.vstack((upperNodes, [n.pos[0], n.pos[1], n.pos[2], n.row]))
            if np.dot(normal, normalRef) < 0:
                for n in e.nodes:
                    if n.row not in lowerNodes[:,3] and n.row not in upperNodes[:,3]:
                        lowerNodes = np.vstack((lowerNodes, [n.pos[0], n.pos[1], n.pos[2], n.row]))

        self.tms['ElmLoop'].stop()

        dummyLower = lowerNodes.copy()
        dummyLower[:,0] *= -1

        invStruct_tr[0] = np.vstack((upperNodes, dummyLower))
        invStruct[0] = np.vstack((upperNodes, lowerNodes))

        # Remove duplicates
        #invStruct_tr[0] = np.unique(invStruct_tr[0], axis=0)

        # Wake part
        for iNode, n in enumerate(blw[1].nodes):
            invStruct_tr[1][iNode, :] = [n.pos[0], n.pos[1], n.pos[2], n.row]
        invStruct[1] = invStruct_tr[1].copy()

        # Create spanwise distribution
        if config['nDim'] == 2:
            spanDistri = np.zeros(1)
            spanDistri[0] = invStruct[0][0,2] # Span direction is the z direction

            wingSpan = 0.
            fInterp = interp1d(invStruct_tr[0][:,0], invStruct_tr[0][:,1])

        elif config['nDim'] == 3:
            invStruct[0][:, [1,2]] = invStruct[0][:, [2, 1]]
            invStruct[1][:, [1,2]] = invStruct[1][:, [2, 1]]
            invStruct_tr[0][:, [1,2]] = invStruct_tr[0][:, [2, 1]]
            invStruct_tr[1][:, [1,2]] = invStruct_tr[1][:, [2, 1]]
            spanDistri = np.linspace(min(invStruct[0][:,2]), max(invStruct[0][:,2]), config['nSections'])
            spanDistri[0] +=0.1
            spanDistri[-1] -=0.1

            # Spline wing
            # Point in the wing tip region are not accounted for during Bezier curve spline

            m = len(invStruct_tr[0][:,0])
            _s = (m-np.sqrt(2*m) + m+np.sqrt(2*m))/2
            tck_wing = bisplrep(invStruct_tr[0][invStruct_tr[0][:,2]<config['span'],0], invStruct_tr[0][invStruct_tr[0][:,2]<config['span'],2], invStruct_tr[0][invStruct_tr[0][:,2]<config['span'],1], kx=5, ky=5, s=1e-6, quiet=1)

        # Create sections
        self.tms['iSecLoop'].start()

        for iSec, z in enumerate(spanDistri):
            portion = invStruct[0][abs(invStruct[0][:,2] - z) <= 0.01 * config['span'],:]
            
            chordLength = max(portion[:, 0]) - min(portion[:, 0])
            xDistri = min(portion[:,0]) + 1/2*(1 + np.cos(np.linspace(0, np.pi, config['nPoints'][0])))*chordLength

            viscStruct[iSec][0][:,0] = np.concatenate((xDistri[:-1], xDistri[::-1]), axis=None)
            viscStruct[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            viscStruct_tr[iSec][0][:,0] = np.concatenate((xDistri[:-1], -xDistri[::-1]), axis=None).copy()
            viscStruct_tr[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            # Interpolation in the transformed space.
            if config['nDim'] == 3:
                    
                for iPoint in range(len(viscStruct_tr[iSec][0][:,0])):
                    viscStruct_tr[iSec][0][iPoint, 1] = bisplev(viscStruct_tr[iSec][0][iPoint,0], viscStruct_tr[iSec][0][iPoint,2], tck_wing)


            elif config['nDim'] == 2:
                viscStruct_tr[iSec][0][:,1] = fInterp(viscStruct_tr[iSec][0][:,0])

            #viscStruct_tr[iSec][0][0,1] = 0.
            #viscStruct_tr[iSec][0][-1,1] = 0.
            viscStruct[iSec][0][:,1] = viscStruct_tr[iSec][0][:,1].copy()

            # Wake 
            wakeLength = np.max(invStruct[1][:,0]) - np.max(viscStruct[iSec][0][:,0])
            viscStruct_tr[iSec][1][:,0] = np.max(viscStruct[iSec][0][:,0]) + (np.max(viscStruct[iSec][0][:,0]) + np.cos(np.linspace(np.pi, np.pi/2, config['nPoints'][1]))) * wakeLength
            viscStruct_tr[iSec][1][:,2] = z*np.ones(len(viscStruct[iSec][1][:,0]))
            viscStruct_tr[iSec][1][:,1] = interp1d(invStruct_tr[1][:,0], invStruct_tr[1][:,1])(viscStruct_tr[iSec][1][:,0])

            viscStruct[iSec][1] = viscStruct_tr[iSec][1].copy()

        # Cg positions
        self.tms['iSecLoop'].stop()

        viscCg_tr = [[np.zeros((len(viscStruct[iSec][iReg][:,0]) - 1, 3)) for iReg in range(2)] for iSec in range(len(viscStruct))]
        invCg_tr = [np.zeros((len(blw[iReg].tag.elems), 3)) for iReg in range(2)]
        self.tms['CgLoop'].start()
        
        # Viscous
        for iSec in range(len(viscStruct)):
            for iReg in range(len(viscStruct[iSec])):
                for iElm in range(len(viscStruct[iSec][iReg][:,0]) - 1):
                    for iDir in range(3):
                        viscCg_tr[iSec][iReg][iElm, iDir] = viscStruct_tr[iSec][iReg][iElm, iDir] + viscStruct_tr[iSec][iReg][iElm + 1, iDir]
                viscCg_tr[iSec][iReg] /= 2

        # Inviscid
        for iReg in range(len(blw)):
            for iElm, e in enumerate(blw[iReg].tag.elems):
                for n in e.nodes:
                    if n.row in upperNodes[:,3] or iReg == 1:
                        invCg_tr[iReg][iElm, 0] +=n.pos[0]
                    elif n.row in lowerNodes[:,3]:
                        invCg_tr[iReg][iElm, 0] += -n.pos[0]

                    invCg_tr[iReg][iElm, 1] +=n.pos[1]
                    invCg_tr[iReg][iElm, 2] +=n.pos[2]
                invCg_tr[iReg][iElm, :] /= e.nodes.size()

            if config['nDim'] == 3:
                invCg_tr[iReg][:,[1,2]] = invCg_tr[iReg][:,[2,1]]
        self.tms['CgLoop'].stop()

        self.invStruct = invStruct
        self.invStruct_tr = invStruct_tr
        self.viscStruct = viscStruct
        self.viscStruct_tr = viscStruct_tr

        self.invCg_tr = invCg_tr
        self.viscCg_tr = viscCg_tr
        """np.set_printoptions(threshold=sys.maxsize)
        for sec in self.viscStruct:
            print('next')
            sec[0][:,[1,2]] = sec[0][:,[2,1]]
            print(sec[0])"""

        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        ax.scatter(invStruct_tr[0][:,0], invStruct_tr[0][:,2], invStruct_tr[0][:,1], s=0.3)
        ax.scatter(invStruct_tr[0][:,0], -invStruct_tr[0][:,2], invStruct_tr[0][:,1], s=0.3, color='grey')
        #ax.scatter(invStruct[1][:,0], invStruct[1][:,2], invStruct[1][:,1], s=0.3)
        for sec in viscStruct_tr:
            ax.scatter(sec[0][:,0], sec[0][:,2], sec[0][:,1], color = 'red')
            ax.scatter(sec[1][:,0], sec[1][:,2], sec[1][:,1], color = 'red')
        ax.set_zlim([-0.5, 0.5])
        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_zlabel('z')
        #plt.gca().invert_yaxis()
        plt.axis('off')
        plt.show()


    def __GetAirfoil(self, blw, config):

        if config['nSections'] != 1:
            raise RuntimeError('Cannot create more than 1 section on 2D geometry. Specified:', config['nSections'])

        viscStruct = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0]-1, 3)) for i in range(2)] for _ in range(config['nSections'])]
        viscStruct_tr = [[np.zeros((config['nPoints'][1] if i==1 else 2*config['nPoints'][0]-1, 3)) for i in range(1)] for _ in range(config['nSections'])]

        # Create node list
        nodesCoord = np.zeros((0, 4))
        for n in blw[0].nodes:
            nodesCoord = np.vstack((nodesCoord, [n.pos[0], n.pos[1], n.pos[2], n.row]))

        spanDistri = np.zeros(1)
        spanDistri[0] = nodesCoord[0,2] # Span direction is the z direction

        wingSpan = max(nodesCoord[:,2]) - min(nodesCoord[:,2])

        #upperNodes = nodesCoord[nodesCoord[:,2]>0]

        """nVec = [np.zeros(0) for _ in range(blw[1].tag.elems[0].nodes.size())]
        nVec = blw[1].tag.elems[0]
        normalRef = np.cross(nVec[1] - nVec[0], nVec[2] - nVec[0])
        normalRef/=np.linalg.norm(normalRef)

        upperNodes = np.zeros((0,4))
        lowerNodes = np.zeros((0,4))
        for e in blw[0].tag.elems:
            # Compute normal
            cVec = [e.nodes[1].pos[0] - e.nodes[0].pos[0], -(e.nodes[1].pos[1] - e.nodes[0].pos[1]), 0]
            print(cVec)
            if np.dot(cVec, normalRef) >=0:
                for n in e.nodes:
                    if n.row not in upperNodes[:,3] and n.row not in lowerNodes[:,3]:
                       upperNodes = np.vstack((upperNodes, [n.pos[0], n.pos[1], n.pos[2], n.row]))

            else:
                for n in e.nodes:
                    if n.row not in upperNodes[:,3] and n.row not in lowerNodes[:,3]:
                       upperNodes = np.vstack((upperNodes, [n.pos[0], n.pos[1], n.pos[2], n.row]))

        plt.plot(upperNodes[:,0], upperNodes[:,1])
        plt.plot(lowerNodes[:,0], lowerNodes[:,1])
        plt.show()"""
        for iSec, z in enumerate(spanDistri):
            
            portion = nodesCoord[abs(nodesCoord[:,2] - z) <= 0.01 * wingSpan,:]

            le_coord = portion[portion[:,0] == np.min((portion[:,0])), :][0]
            te_coord = portion[portion[:,0] == np.max((portion[:,0])), :][0]

            # Find te elements
            te_elems = []
            for e in blw[0].tag.elems:
                for n in e.nodes:
                    if n.pos[0] == te_coord[0]:
                        te_elems.append(e)
                        break
            
            # Cg of te elements
            cg_teElems = np.zeros((len(te_elems), 3))
            for iElm, eTe in enumerate(te_elems):
                for n in eTe.nodes:
                    cg_teElems[iElm, 0] += n.pos[0]
                    cg_teElems[iElm, 1] += n.pos[1]
                    cg_teElems[iElm, 2] += n.pos[2]
            
            # Sort upper side in 0 and lower side in 1
            if cg_teElems[0, 1] < cg_teElems[1, 1]:
                save = te_elems[0].copy()
                te_elems[0] = te_elems[1]
                te_elems[1] = save
            
            # Find te_nodes (sorted [upper, lower])
            te_nodes = [None, None]
            for i in range(len(te_elems)):
                for n in te_elems[i].nodes:
                    if n.pos[0] == te_coord[0]:
                        te_nodes[i] = n
                        break
            
            # Find le_node
            for n in blw[0].nodes:
                if n.pos[0] == le_coord[0]:
                    le_node = n
                    break
                

            meany = (le_coord[1] + te_coord[1]) / 2

            transformedCoord = np.zeros((0, 4))
            nodesList = []

            for iNode, n in enumerate(blw[0].nodes):
                if n.pos[0] != le_coord[0] and n.pos[0] != te_coord[0]:
                    if n.pos[1] > meany:
                        transformedCoord = np.vstack((transformedCoord, [n.pos[0], n.pos[1], n.pos[2], n.row]))
                        nodesList.append(n)

                    elif n.pos[1] < meany:
                        transformedCoord = np.vstack((transformedCoord, [-n.pos[0], n.pos[1], n.pos[2], n.row]))
                        nodesList.append(n)

            transformedCoord = np.vstack((transformedCoord, le_coord))
            nodesList.append(le_node)
            transformedCoord = np.vstack((transformedCoord, [te_nodes[0].pos[0], te_nodes[0].pos[1], te_nodes[0].pos[2], te_nodes[0].row]))
            nodesList.append(te_nodes[0])
            transformedCoord = np.vstack((transformedCoord, [-te_nodes[1].pos[0], te_nodes[1].pos[1], te_nodes[1].pos[2], te_nodes[1].row]))
            nodesList.append(te_nodes[1])

            wakeCoord = np.zeros((len(blw[1].nodes), 4))
            for iNode, n in enumerate(blw[1].nodes):
                wakeCoord[iNode, 0] = n.pos[0]
                wakeCoord[iNode, 1] = n.pos[1]
                wakeCoord[iNode, 2] = n.pos[2]
                wakeCoord[iNode, 3] = n.row
            
            chordLength = max(portion[:, 0]) - min(portion[:, 0])
            xDistri = le_coord[0] + 1/2*(1 + np.cos(np.linspace(0, np.pi, config['nPoints'][0])))*chordLength

            viscStruct[iSec][0][:,0] = np.concatenate((xDistri[:-1], xDistri[::-1]), axis=None)
            viscStruct[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            viscStruct_tr[iSec][0][:,0] = np.concatenate((xDistri[:-1], -xDistri[::-1]), axis=None).copy()
            viscStruct_tr[iSec][0][:,2] = z*np.ones(len(viscStruct[iSec][0][:,0]))

            # Interpolation in the transformed space.
            splineTr = transformedCoord.copy()
            splineTr = splineTr[splineTr[:, 0].argsort()]
            spl = splrep(splineTr[:,0], splineTr[:,1], k=5)
            viscStruct_tr[iSec][0][:,1] = splev(viscStruct_tr[iSec][0][:,0], spl)
            #viscStruct_tr[iSec][0][:,1] = interp1d(transformedCoord[:,0], transformedCoord[:,1])(viscStruct_tr[iSec][0][:,0])
            viscStruct[iSec][0][:,1] = viscStruct_tr[iSec][0][:,1].copy()

            # Wake
            wakeLength = np.max(wakeCoord[:,0]) - np.max(viscStruct[iSec][0][:,0])
            viscStruct[iSec][1][:,0] = np.max(viscStruct[iSec][0][:,0]) + (np.max(viscStruct[iSec][0][:,0]) + np.cos(np.linspace(np.pi, np.pi/2, config['nPoints'][1]))) * wakeLength
            viscStruct[iSec][1][:,2] = z*np.ones(len(viscStruct[iSec][1][:,0]))
            viscStruct[iSec][1][:,1] = interp1d(wakeCoord[:,0], wakeCoord[:,1])(viscStruct[iSec][1][:,0])

            self.invStruct = [np.zeros((0, 4)) for _ in range(2)]
            self.invStruct_tr = [np.zeros((0, 4)) for _ in range(2)]
            self.invStruct[0] = abs(transformedCoord.copy())
            self.invStruct_tr[0] = transformedCoord.copy()
            self.invStruct[1] = abs(wakeCoord.copy())
            self.invStruct_tr[1] = wakeCoord.copy()

            print(np.shape(self.invStruct[0]))


            self.viscStruct = viscStruct
            self.viscStruct_tr = viscStruct_tr
            #plt.plot(nodesCoord[:,0], nodesCoord[:,1], 'o', markersize=8, color='green')
            """plt.plot(viscStruct[0][0][:,0], viscStruct[0][0][:,1], 'x', color='blue')
            plt.axis('off')
            plt.show()"""
            return nodesList, transformedCoord, wakeCoord, viscStruct, viscStruct_tr

    def GetMesh(self, msh, visc=0):
        """Obtain the nodes' location and row and cg of all elements.
        If the mesh is the viscous one, TE nodes are duplicated (by creating a new row).
        They exist for the interpolation and for the solution computation.
        If not, the nodes are pseudo-duplicated (without changing the row). They therefore
        exist only for the interpolation but not for the solution."""

        data = [np.zeros((0,4)) for _ in range(2)]
        cg = [np.zeros((0,4)) for _ in range(2)]

        nodesAttTe = []

        for e in msh.elems:
            # Compute cg position
            cg_pt = np.zeros(3)
            for n in e.nodes:
                cg_pt += [n.pos[0], n.pos[1], n.pos[2]]
            if visc and e.nodes.size() == 2:
                cg_pt = None
            else:
                cg_pt/=e.nodes.size()
                cg_pt = np.hstack((cg_pt, e.no))

            if e.ptag.name == "te":
                for nw in e.nodes:
                    nodesAttTe.append(nw.row)
                if cg_pt is not None:
                    cg[0] = np.vstack((cg[0], cg_pt))
            elif e.ptag.name == "wake":
                for nw in e.nodes:
                    if nw.row not in data[1][:,3]:
                        data[1] = np.vstack((data[1], [nw.pos[0], nw.pos[1], nw.pos[2], nw.row]))
                if cg_pt is not None:
                    cg[1] = np.vstack((cg[1], cg_pt))
            elif e.ptag.name == 'wing':
                for nw in e.nodes:
                    if nw.row not in data[0][:,3]:
                        data[0] = np.vstack((data[0], [-nw.pos[0], nw.pos[1], nw.pos[2], nw.row]))
                if cg_pt is not None:
                    cg_pt[0] = -cg_pt[0]
                    cg[0] = np.vstack((cg[0], cg_pt))
            elif e.ptag.name == 'wing_':
                for nw in e.nodes:
                    if nw.row not in data[0][:,3]:
                        data[0] = np.vstack((data[0], [nw.pos[0], nw.pos[1], nw.pos[2], nw.row]))
                if cg_pt is not None:
                    cg[0] = np.vstack((cg[0], cg_pt))
        
        val, counts = np.unique(nodesAttTe, return_counts=True)
        teNodes = []
        for i, c in enumerate(counts):
            if c==2:
                teNodes.append(val[i])
            elif c!=1:
                raise RuntimeError('Found TE node more than 2 times', val[i])
        
        maxRowNb = max(data[0][:,3])
        currMax = data[0][0,:].copy()
        currMin = data[0][0,:].copy()
        for i, r in enumerate(data[0][:,3]):
            if data[0][i,1] > currMax[1] or data[0][i,0] < currMax[0]:
                currMax = data[0][i,:].copy()
            if data[0][i,1] < currMin[1] or data[0][i,0] < currMin[0]:
                currMin = data[0][i,:].copy()
            if r in teNodes:
                if visc:
                    data[0] = np.vstack((data[0], [-data[0][i,0], data[0][i,1], data[0][i,2], data[0][i,3]+maxRowNb]))
                else:
                    data[0] = np.vstack((data[0], [-data[0][i,0], data[0][i,1], data[0][i,2], data[0][i,3]]))
        currMax[0] *=-1
        currMin[0] *=-1
        if visc:
            currMax[3] += maxRowNb
            currMin[3] += 2*maxRowNb
        data[0] = np.vstack((data[0], currMax))
        data[0] = np.vstack((data[0], currMin))

        return data, cg

    def SortMesh(self, data, cg):

        data = data.copy()

        data[0] = data[0][data[0][:,1]<=1, :]

        cg = cg.copy()

        cg[0] = cg[0][cg[0][:,1]<=1, :]

        sections = data[0][:,1].copy()
        for i in range(len(sections)):
            sections[i] = round(sections[i], 3)
        sections = np.unique(sections)
        sections = np.sort(sections)

        nChord = [len(data[i][abs(data[i][:,1]-sections[0])<1e-3, 0]) for i in range(len(data))]

        sortedRows = [np.zeros((nChord[i], 0)) for i in range(len(data))]
        Xp = [np.zeros((nChord[i], 0)) for i in range(len(data))]
        Yp = [np.zeros((nChord[i], 0)) for i in range(len(data))]
        Zp = [np.zeros((nChord[i], 0)) for i in range(len(data))]

        for iReg, reg in enumerate(data):

            for iy, y in enumerate(sections):

                # Find nearest value in the mesh

                idx = (np.abs(reg[:,1]-y).argmin())
                sections[iy] = reg[idx,1]

                sort = reg[abs(reg[:,1]-sections[iy])<1e-3, :]
                sort = sort[sort[:,0].argsort()]

                sortedRows[iReg] = np.column_stack((sortedRows[iReg], sort[:,3]))
                Xp[iReg] = np.column_stack((Xp[iReg], sort[:,0]))
                Yp[iReg] = np.column_stack((Yp[iReg], sort[:,1]))
                Zp[iReg] = np.column_stack((Zp[iReg], sort[:,2]))

        sections_cg = cg[0][:,1].copy()
        for i in range(len(sections_cg)):
            sections_cg[i] = round(sections_cg[i], 3)
        sections_cg = np.unique(sections_cg)
        sections_cg = np.sort(sections_cg)

        nChord_cg = [len(cg[i][abs(cg[i][:,1]-cg[i][0,1])<1e-3, 0]) for i in range(len(cg))]

        Xc = [np.zeros((nChord_cg[i], 0)) for i in range(len(data))]
        Yc = [np.zeros((nChord_cg[i], 0)) for i in range(len(data))]
        Zc = [np.zeros((nChord_cg[i], 0)) for i in range(len(data))]

        sortedElems = [np.zeros((nChord_cg[i], 0)) for i in range(len(cg))]
        
        for iReg, reg in enumerate(cg):

            for iy, y in enumerate(sections_cg):

                # Find nearest value in the mesh

                idx = (np.abs(reg[:,1]-y).argmin())
                sections_cg[iy] = reg[idx,1]

                sort = reg[abs(reg[:,1]-sections_cg[iy])<1e-3, :]
                sort = sort[sort[:,0].argsort()]
                
                sortedElems[iReg] = np.column_stack((sortedElems[iReg], sort[:,3]))
                Xc[iReg] = np.column_stack((Xc[iReg], sort[:,0]))
                Yc[iReg] = np.column_stack((Yc[iReg], sort[:,1]))
                Zc[iReg] = np.column_stack((Zc[iReg], sort[:,2]))

        return Xp, Yp, Zp, sortedRows, Xc, Yc, Zc, sortedElems

        



            

            