#ifndef WBLREGION_H
#define WBLREGION_H

#include "vii.h"
#include "wClosures.h"
#include "wViscMesh.h"
#include "wInvBnd.h"

namespace vii
{

class DART_API BLRegion
{

private:
    unsigned int const nVar = 5;    /* Number of variables of the partial differential problem */

    /* Transition related variables */

    double nCrit = 9.0;             /* Critical amplification factor */


public:

    ViscMesh *mesh;                 /* Boundary layer mesh */
    InvBnd *invBnd;                 /* Inviscid quantities at the edge of the boundary layer */
    Closures *closSolver;           /* Closure relatrions module */

    std::string name;               /* Name of the region */

    /* Boundary layer variables */

    std::vector<double> U;          /* Solution vector (theta, H, N, ue, Ct)                          */
    std::vector<double> Ret;        /* Reynolds number based on the momentum thickness (theta)        */
    std::vector<double> Cd;         /* Local dissipation coefficient                                  */
    std::vector<double> Cf;         /* Local friction coefficient                                     */
    std::vector<double> Hstar;      /* Kinetic energy shape parameter (thetaStar/theta) */
    std::vector<double> Hstar2;     /* Density shape parameter (deltaStarStar/theta) */
    std::vector<double> Hk;         /* Kinematic shape parameter (int(1-u/ue deta)) */
    std::vector<double> Cteq;       /* Equilibrium shear stress coefficient (turbulent BL) */
    std::vector<double> Us;         /* Equivalent normalized wall slip velocity */
    std::vector<double> Delta;      /* Boundary layer thickness */

    std::vector<double> DeltaStar;  /* Dispacement thickness (int(1-rho*u/rhoe*ue deta)) */

    /* Transition related variables */

    std::vector<int> Regime;        /* Laminar (0) or turbulent (1) regime */
    std::vector<double> Blowing;    /* Blowing velocity */

    unsigned int transMarker;       /* Marker of the transition location */
    double xtrF;                    /* Forced transition location */
    double xtr;                     /* Transition location (=xtrF if forced) */

    BLRegion(double _xtrF = -1.0);
    ~BLRegion();

    /* Distribute BC */
    void SetMesh(std::vector<double> x,
                 std::vector<double> y,
                 std::vector<double> z);
    
    /* Boundary conditions */
    void SetStagBC(double Re);
    void SetWakeBC(double Re, std::vector<double> UpperCond, std::vector<double> LowerCond);
    
    /* General access */
    unsigned int GetnVar() const {return nVar;};
    double GetnCrit() const {return nCrit;};

    /* Others */
    void PrintSolution(size_t iPoint) const;
    void ComputeBlwVel();

};
} // namespace dart
#endif //WBLREGION_H
