/* ----------------------------------------------------------------------------------- */
/*                                                                                     */
/*                       Coupled integral boundary layer solver                        */
/*                                                                                     */
/*  Université de Liège                                                                */
/*  February 2022                                                                      */
/*  Author: Paul Dechamps                                                              */
/*  v0.1                                                                               */
/* ----------------------------------------------------------------------------------- */

/* --- Project includes -------------------------------------------------------------- */

#include "wViscMesh.h"
#include <cmath>

/* --- Namespaces -------------------------------------------------------------------- */

using namespace vii;

/* ----------------------------------------------------------------------------------- */

ViscMesh::ViscMesh(){nMarkers = 0; nElm = 0; prevnMarkers = 0;}

ViscMesh::~ViscMesh(){}

void ViscMesh::SetGlob(std::vector<double> _x, std::vector<double> _y, std::vector<double> _z)
{
  if (_x.size() != _y.size() || _y.size() != _z.size() || _x.size() != _z.size())
  {
    throw std::runtime_error("vii::ViscMesh Wrong mesh sizes.");
  }

  nMarkers = _x.size();
  x.resize(nMarkers, 0.);
  y.resize(nMarkers, 0.);
  z.resize(nMarkers, 0.);
  nElm = nMarkers - 1;
  x = _x;
  y = _y;
  z = _z;

  ComputeBLcoord();
}

void ViscMesh::ComputeBLcoord()
{
    Loc.resize(nMarkers, 0.);
    alpha.resize(nElm, 0.);
    dx.resize(nElm, 0.);

    for (size_t iPoint=0; iPoint<nElm; ++iPoint)
    {
      alpha[iPoint] = std::atan2(y[iPoint+1]-y[iPoint], x[iPoint+1]-x[iPoint]);
      /* dx = sqrt(Delta x ^2 + Delta y ^2) */
      dx[iPoint] = std::sqrt((x[iPoint+1]-x[iPoint])*(x[iPoint+1]-x[iPoint]) + (y[iPoint+1]-y[iPoint])*(y[iPoint+1]-y[iPoint]));
      Loc[iPoint + 1] = Loc[iPoint] + dx[iPoint];
    }
}
