#ifndef WVISCSOLVER_H
#define WVISCSOLVER_H

#include "vii.h"
#include "wBLRegion.h"
#include "wTimers.h"
namespace vii
{

class DART_API ViscSolver
{
public:
    double Cdt,
           Cdf,
           Cdp;
    
    std::vector<double> Cdt_sec,
                        Cdf_sec,
                        Cdp_sec;

    std::vector<std::vector<BLRegion *>> Sections;

    unsigned int verbose;

private:
    double Re,
           CFL0;
    
    int solverExitCode;

    std::vector<bool> stagPtMvmt;

    TimeSolver *tSolver;

    double Span; /* Wing Span (0 if 2D case) */

protected:
    fwk::Timers tms;       ///< internal timers

public:
    ViscSolver(double _Re, double _Minf, double _CFL0, unsigned int nSections, double _Span=0., unsigned int verbose=1);
    ~ViscSolver();
    int Run(unsigned int const couplIter);
    double GetRe() const {return Re;};

    void SetGlobMesh(size_t iSec, size_t iRegion, std::vector<double> _x, std::vector<double> _y, std::vector<double> _z);
    void SetVelocities(size_t iSec, size_t iRegion, std::vector<double> _vx, std::vector<double> _vy, std::vector<double> _vz);
    void SetMe(size_t iSec, size_t iRegion, std::vector<double> _Me);
    void SetRhoe(size_t iSec, size_t iRegion, std::vector<double> _Rhoe);
    void SetxxExt(size_t iSec, size_t iRegion, std::vector<double> _xxExt);
    void SetDeltaStarExt(size_t iSec, size_t iRegion, std::vector<double> _DeltaStarExt);

    std::vector<double> ExtractBlowingVel(size_t iSec, size_t iRegion) const;
    std::vector<double> Extractxx(size_t iSec, size_t iRegion) const;
    std::vector<double> ExtractDeltaStar(size_t iSec, size_t iRegion) const;
    std::vector<double> ExtractxCoord(size_t iSec, size_t iRegion) const;

    std::vector<double> ExtractUe(size_t iSec, size_t iRegion) const;
    
    void SetUeOld(size_t iSec, size_t iRegion, std::vector<double> _ue);


    double Getxtr(size_t iSec, size_t iRegion) const {return Sections[iSec][iRegion]->xtr;};

    std::vector<std::vector<double>> GetSolution(size_t iSec);


private:
    void CheckWaves(size_t iPoint, BLRegion *bl);
    void AverageTransition(size_t iPoint, BLRegion *bl, int forced);
    void ComputeSectionalDrag(std::vector<BLRegion *> const &bl, int i);
    void ComputeTotalDrag();
    void ComputeBlwVel();
    void PrintBanner() const;
};
} // namespace dart
#endif //WVISCSOLVER_H